"""
Parser for bash statements using tree-sitter.

This module provides functionality to parse and identify individual bash statements.
"""

from .bash_statement_parser import BashStatementParser, Statement
